#######################################################################################################################################################################################################
#
# Changes made to this constructor
# General
#     2021-Sep-30: move this function out of BrooksCorey struct as an external method for the constructor (avoid dependency on ConstrainedRootSolvers)
#     2022-Apr-19: fix documentation
#
#######################################################################################################################################################################################################
"""

    BrooksCorey{FT}(vg::VanGenuchten{FT}) where {FT<:AbstractFloat}

A constructor for BrooksCorey to create BrooksCorey type soil from VanGenuchten type, given
- `vg` `VanGenuchten` type soil water retention curve
"""
BrooksCorey{FT}(vg::VanGenuchten{FT}) where {FT<:AbstractFloat} = (
    # generate data to fit
    _Θs   = range(vg.Θ_RES+FT(1e-2); stop=vg.Θ_SAT-FT(1e-2), length=30);
    _Ψ_vG = -1 .* soil_ψ_25.([vg], _Θs);

    # function to fit BrooksCorey parameters
    @inline _fit(x) = (
        _bc   = BrooksCorey{FT}(vg.TYPE, x[1], x[2], vg.Θ_SAT, vg.Θ_RES);
        _Ψ_BC = -1 .* soil_ψ_25.([_bc], _Θs);
        _diff = sum( (log.(_Ψ_BC) .- log.(_Ψ_vG)) .^ 2 );
        return -_diff
    );

    _st  = SolutionToleranceND{FT}([1e-3, 1e-6], 30);
    _ms  = ReduceStepMethodND{FT}(x_mins = FT[1e-3, 1e-6], x_maxs = FT[ 100, 1000], x_inis = [(2*vg.N-1) / (vg.N-1), 1 / (vg.Α)], Δ_inis = FT[0.1, 1e-3]);
    _sol = find_peak(_fit, _ms, _st);

    return BrooksCorey{FT}(vg.TYPE, _sol[1], _sol[2], vg.Θ_SAT, vg.Θ_RES);
);


#######################################################################################################################################################################################################
#
# Changes made to this function
# General
#     2021-Sep-30: create this function to work with two soil types using either VanGenuchten or BrooksCorey function
#
#######################################################################################################################################################################################################
"""
This function calculates soil metric potential from soil retention curve type and soil volumetric water potential. The supported methods are

$(METHODLIST)

"""
function soil_ψ_25 end


#######################################################################################################################################################################################################
#
# Changes made to this method
# General
#     2021-Apr-19: add documentation
#
#######################################################################################################################################################################################################
"""

    soil_ψ_25(bc::BrooksCorey{FT}, θ::FT) where {FT<:AbstractFloat}

Return the soil metric potential, given
- `bc` `BrooksCorey` type structure
- `θ` Soil volumetric water content (absolute value)
"""
soil_ψ_25(bc::BrooksCorey{FT}, θ::FT) where {FT<:AbstractFloat} = (
    @unpack B, Ψ_SAT, Θ_RES, Θ_SAT = bc;

    # calculate effective θ
    _θ_e = min(1, max(0, (θ - Θ_RES) / (Θ_SAT - Θ_RES)));

    # if _θ_e >= 1, return 0
    if _θ_e >= 1
        return FT(0)
    end;

    return -Ψ_SAT / (_θ_e ^ B)
);


#######################################################################################################################################################################################################
#
# Changes made to this method
# General
#     2021-Apr-19: add documentation
#
#######################################################################################################################################################################################################
"""

    soil_ψ_25(vg::VanGenuchten{FT}, θ::FT) where {FT<:AbstractFloat}

Return the soil metric potential, given
- `vg` `VanGenuchten` type structure
- `θ` Soil volumetric water content (absolute value)
"""
soil_ψ_25(vg::VanGenuchten{FT}, θ::FT) where {FT<:AbstractFloat} = (
    @unpack M, N, Α, Θ_RES, Θ_SAT = vg;

    # calculate effective θ
    _θ_e = min(1, max(0, (θ - Θ_RES) / (Θ_SAT - Θ_RES)));

    # if _θ_e >= 1, return 0
    if _θ_e >= 1
        return FT(0)
    end;

    return -1 * (_θ_e ^ (-1/M) - 1) ^ (1/N) / Α
);


#######################################################################################################################################################################################################
#
# Changes to the function
# General
#     2022-Feb-01: migrate function to version v0.3 of PlantHydraulics
#     2022-Feb-01: add documentation
#     2022-Feb-01: rename the function to relative_hydraulic_conductance
#     2022-Feb-01: remove viscosity correction
#     2022-Apr-19: move function to SoilHydraulics from PlantHydraulics (will be imported in PlantHydraulics)
#
#######################################################################################################################################################################################################
"""
This function return the hydraulic conductance ralative to maximum. The supported methods are

$(METHODLIST)

"""
function relative_hydraulic_conductance end


#######################################################################################################################################################################################################
#
# Changes to the method
# General
#     2022-Apr-19: move method to SoilHydraulics from PlantHydraulics
#     2022-Apr-19: add documentation
#
#######################################################################################################################################################################################################
"""

    relative_hydraulic_conductance(bc::BrooksCorey{FT}, θ::FT) where {FT<:AbstractFloat}

Return the relative hydraulic conductance of the soil, given
- `bc` `BrooksCorey` type structure
- `θ` Soil volumetric water content (absolute value)
"""
relative_hydraulic_conductance(bc::BrooksCorey{FT}, θ::FT) where {FT<:AbstractFloat} = (
    @unpack B, Θ_RES, Θ_SAT = bc;

    _θ_e = min(1, max(0, (θ - Θ_RES) / (Θ_SAT - Θ_RES)));

    return _θ_e ^ (2 * B + 3)
);


#######################################################################################################################################################################################################
#
# Changes to the method
# General
#     2022-Apr-19: move method to SoilHydraulics from PlantHydraulics
#     2022-Apr-19: add documentation
#
#######################################################################################################################################################################################################
"""

    relative_hydraulic_conductance(bc::BrooksCorey{FT}, ψ::Bool, ψ_25::FT) where {FT<:AbstractFloat}

Return the relative hydraulic conductance of the soil, given
- `bc` `BrooksCorey` type structure
- `ψ` Bool to indicate that next parameter is potential
- `ψ_25` Soil metric potential at a reference temperature of 25 °C
"""
relative_hydraulic_conductance(bc::BrooksCorey{FT}, ψ::Bool, ψ_25::FT) where {FT<:AbstractFloat} = (
    @unpack B, Ψ_SAT = bc;

    _θ_e = (-Ψ_SAT / ψ_25) ^ (1 / B);

    return _θ_e ^ (2 * B + 3)
);


#######################################################################################################################################################################################################
#
# Changes to the method
# General
#     2022-Apr-19: move method to SoilHydraulics from PlantHydraulics
#     2022-Apr-19: add documentation
#
#######################################################################################################################################################################################################
"""

    relative_hydraulic_conductance(vg::VanGenuchten{FT}, θ::FT) where {FT<:AbstractFloat}

Return the relative hydraulic conductance of the soil, given
- `vg` `VanGenuchten` type structure
- `θ` Soil volumetric water content (absolute value)
"""
relative_hydraulic_conductance(vg::VanGenuchten{FT}, θ::FT) where {FT<:AbstractFloat} = (
    @unpack M, N, Α, Θ_RES, Θ_SAT = vg;

    _θ_e = min(1, max(0, (θ - Θ_RES) / (Θ_SAT - Θ_RES)));

    return sqrt(_θ_e) * (1 - (1 - _θ_e ^ (1 / M)) ^ M)^2
);


#######################################################################################################################################################################################################
#
# Changes to the method
# General
#     2022-Apr-19: move method to SoilHydraulics from PlantHydraulics
#     2022-Apr-19: add documentation
#
#######################################################################################################################################################################################################
"""

    relative_hydraulic_conductance(vg::VanGenuchten{FT}, ψ::Bool, ψ_25::FT) where {FT<:AbstractFloat}

Return the relative hydraulic conductance of the soil, given
- `vg` `VanGenuchten` type structure
- `ψ` Bool to indicate that next parameter is potential
- `ψ_25` Soil metric potential at a reference temperature of 25 °C
"""
relative_hydraulic_conductance(vg::VanGenuchten{FT}, ψ::Bool, ψ_25::FT) where {FT<:AbstractFloat} = (
    @unpack M, N, Α = vg;

    _θ_e = (1 / (1 + (-ψ_25 * Α) ^ N)) ^ M;

    return sqrt(_θ_e) * (1 - (1 - _θ_e ^ (1 / M)) ^ M)^2
);
